{------------------------------------------------------------------------------}
{ TWAIN Sample Source                                                          }
{ File: caps.c                                                                 }
{                                                                              }
{ This source code was converted from C to Turbo Pascal for Windoes by         }
{                                                                              }
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

{------------------------------------------------------------------------------}
{$A+} { Align data Byte = $A-, Word = $A+.                                     }
{$F+} { Force FAR call.                                                        }
{$G+} { Generate 80286 code.                                                   }
{$N+} { CoProcessor code generation.                                           }
{$R-} { Range check.                                                           }
{$W-} { Protected mode only $W-, Real mode stack frame support $W+.            }
{$X+} { Extended syntax (When on, function can be used as procedures).         }
{------------------------------------------------------------------------------}

unit twd_caps;

interface

uses WinTypes,
     twain;

function ResetValues                (CapStr    : PChar) : TW_UINT16;

function SetValues                  (pData     : pTW_CAPABILITY;
                                     Msg       : TW_UINT16;
                                     CapStr    : PChar) : TW_UINT16;

function GetIndex                   (Msg       : integer;
                                     CapStr    : PChar) : integer;

function SetNewCap                  (pData     : pTW_CAPABILITY;
                                     CapStr    : PChar) : TW_INT16;

function IsCapAllowed               (Cap       : TW_INT16;
                                     CapStr    : PChar;
                                     Attribute : PChar) : bool;

function WithinRange                (Cap       : TW_UINT16;
                                     CapStr    : PChar) : bool;

procedure ConvertAttributeToText    (Attrib    : TW_UINT32;
                                     StrText   : PChar;
                                     CapStr    : PChar);

function RangeMinimum               (Attr      : PChar;
                                     ItemType  : TW_UINT16) : TW_UINT32;

function RangeMaximum               (Attr      : PChar;
                                     ItemType  : TW_UINT16) : TW_UINT32;

function RangeStep                  (Attr      : PChar;
                                     ItemType  : TW_UINT16) : TW_UINT32;

{$IFNDEF WIN32}
procedure GetPrivateProfileSection16(CapStr    : PChar;
                                     LineOut   : PChar;
                                     Length    : integer);
{$ENDIF}

function FloatToFIX32               (floater   : real) : TW_FIX32;

function FIX32ToFloat               (fix32     : TW_FIX32) : real;


implementation

uses WinProcs, SysUtils, {Win31,    strings, }
     twd_type, twd_prot, twd_com;

var Destination : array[0..50] of char;
    SaveToken   : array[0..511] of char;


function atoi(a : PChar) : longint;
var s    : array[0..31] of char;
    i    : longint;
    Code : integer;
begin
  StrCopy(s, a);
  Val(s, i, Code);
  if (Code <> 0)
  then i := 0;
  atoi := i;
end; { End atoi.                                                               }


function atof(a : PChar) : double;
var s    : array[0..31] of char;
    d    : double;
    Code : integer;
begin
  StrCopy(s, a);
  Val(s, d, Code);
  if (Code <> 0)
  then d := 0.0;
  atof := d;
end; { End atof.                                                               }


function StrTok(a : PChar; b : PChar; c : PChar) : PChar;
var Token  : array[0..512] of char;
    pToken : PChar;
begin
  if (a <> Nil)
  then begin
       SaveToken[0] := #0;
       StrLCopy(Token, a, 512);
       pToken := StrScan(Token, b[0]);
       if (pToken <> Nil)
       then begin
            StrCopy(SaveToken, pToken + 1);
            pToken[0] := #0;
            StrCopy(c, Token);
       end
       else begin
            StrCopy(c, Token);
            SaveToken[0] := #0;
       end;
  end
  else begin
       StrLCopy(Token, SaveToken, 512);
       pToken := StrScan(Token, b[0]);
       if (pToken <> Nil)
       then begin
            StrCopy(SaveToken, pToken + 1);
            pToken[0] := #0;
            StrCopy(c, Token);
       end
       else begin
            StrCopy(c, Token);
            SaveToken[0] := #0;
       end;
  end;
  StrTok := c;
end; { End StrTok.                                                             }


{------------------------------------------------------------------------------}
{ ResetValues -- Set the current values in the source control to the default   }
{ entry and return the Cap structure to the app.                               }
{ In all cases the allowed values are also reset if needed or not (to PowerOn  }
{ settings).  Note that a container type                                       }
{ of range would do this.                                                      }
{------------------------------------------------------------------------------}

function ResetValues(CapStr : PChar) : TW_UINT16;
var Capabilities : array[0..80] of char;
    StrReset     : array[0..80] of char;
begin
  { Ensure that the operation is legal.                                        }
  GetPrivateProfileString(CapStr, 'Reset', '', StrReset, SizeOf(StrReset), SourceControl);
  if (StrLComp(StrReset, 'TWCC_BADPROTOCOL', StrLen('TWCC_BADPROTOCOL')) = 0)
  then begin
       ResetValues := TWCC_BADPROTOCOL;
       exit;
  end;

  lstrcpyn(destination, 'Current', SizeOf(destination));
  if ((StrLComp(StrReset, 'TWON_RANGE',       StrLen('TWON_RANGE'))       = 0) or
      (StrLComp(StrReset, 'TWON_ENUMERATION', StrLen('TWON_ENUMERATION')) = 0))
  then LStrCpyn(destination, 'Allowed', SizeOf(destination));

  GetPrivateProfileString  (CapStr, 'PowerON', '', Capabilities, SizeOf(Capabilities), SourceControl);
  WritePrivateProfileString(CapStr, 'Allowed',     Capabilities, SourceControl);
  GetPrivateProfileString  (CapStr, 'Default', '', Capabilities, SizeOf(Capabilities), SourceControl);
  WritePrivateProfileString(CapStr, 'Current',     Capabilities, SourceControl);

  ResetValues := TWCC_SUCCESS;
end; { End ResetValues.                                                        }


{------------------------------------------------------------------------------}
{ SetValues -- Packs the container structure based first on the type           }
{ of container as listed in the .src file and secondly on the ammount          }
{ and type of data present in the list                                         }
{------------------------------------------------------------------------------}

function SetValues(pData  : pTW_CAPABILITY;
                   Msg    : TW_UINT16;
                   CapStr : PChar) : TW_UINT16;
var Buffer       : array[0..80] of char;
    Capabilities : array[0..80] of char;
    Container    : array[0..120] of char;
    ConType      : integer;
    DataType     : array[0..80] of char;
    wValue       : word;
    lValue       : longint;
    pValue       : pointer;
    ValueType    : TW_UINT16;
    TwackerIni   : array[0..120] of char;
    twEnum       : TW_ENUMERATION;
    twRange      : TW_RANGE;
    twArray      : TW_ARRAY;
    Token        : array[0..256] of char;
    fix32        : TW_FIX32;
    Types        : array[0..9] of byte;
    i            : integer;
begin
  SetValues := TWRC_SUCCESS;
  for i := 0 to 9
  do Types[i] := 99;

  { Make the inifile path.                                                     }
  GetWindowsDirectory(TwackerIni, SizeOf(TwackerIni));
  lstrcat(TwackerIni, '\twacker.ini');

  { Determine the Container type needed and which attribute to set, either     }
  { current, allowed, or default                                               }
  case Msg of
  MSG_GET        : begin
                     GetPrivateProfileString(CapStr, 'Get', '',
                                             Container,
                                             SizeOf(Container),
                                             SourceControl);
                     if (lstrcmp(Container, 'TWON_ONEVALUE') = 0)
		     then GetPrivateProfileString(CapStr, 'Current', 'NONE',
                                                  Capabilities,
                                                  SizeOf(Capabilities),
                                                  SourceControl)
		     else GetPrivateProfileString(CapStr, 'Allowed', 'NONE',
                                                  Capabilities,
                                                  SizeOf(Capabilities),
                                                  SourceControl);
                   end;
  MSG_GETDEFAULT : begin
                     GetPrivateProfileString(CapStr, 'Default', 'NONE',
                                             Capabilities,
                                             SizeOf(Capabilities),
                                             SourceControl);
                     GetPrivateProfileString(CapStr, 'GetDefault', '',
                                             Container,
                                             SizeOf(Container),
                                             SourceControl);
                   end;
  MSG_GETCURRENT : begin
                     GetPrivateProfileString(CapStr, 'Current', 'NONE',
                                             Capabilities,
                                             SizeOf(Capabilities),
                                             SourceControl);
                     GetPrivateProfileString(CapStr, 'GetCurrent', '',
                                             Container,
                                             SizeOf(Container),
                                             SourceControl);
                   end;
  MSG_RESET      : begin
		     GetPrivateProfileString(CapStr, 'Reset', '',
                                             Container,
                                             SizeOf(Container),
                                             SourceControl);
                     if (lstrcmp(Container, 'TWON_ONEVALUE') = 0)
            	     then GetPrivateProfileString(CapStr, 'Current', 'NONE',
                                                  Capabilities,
                                                  SizeOf(Capabilities),
                                                  SourceControl)
                     else GetPrivateProfileString(CapStr, 'Allowed', 'NONE',
                                                  Capabilities,
                                                  SizeOf(Capabilities),
                                                  SourceControl);
                   end;
  MSG_SET        : begin
		     GetPrivateProfileString(CapStr, Destination, 'NONE',
                                             Capabilities,
                                             SizeOf(Capabilities),
                                             SourceControl);
                     GetPrivateProfileString(CapStr, 'Set', '',
                                             Container,
                                             SizeOf(Container),
                                             SourceControl);
                   end;
  end;

  { Notify that something went wrong, probably a bad or unsupported Capability }
  if (lstrcmp(Capabilities, 'NONE') = 0)
  then begin
       StrCopy(Buffer, 'Unsupported Capability. ');
       StrCat(Buffer, CapStr);
       {$IFDEF DEBUG}
         MessageBox(0, Buffer, 'Sample Source', MB_OK);
       {$ENDIF}
       SetValues := TWCC_BADCAP;
  end;

  { Determine the ItemType for the data                                        }
  GetPrivateProfileString(CapStr, 'Type', '', DataType, SizeOf(DataType), SourceControl);
  ValueType := TW_UINT16(GetPrivateProfileInt('ItemType', DataType, 99, TwackerIni));

  { Check type of container needed.  The TWAIN spec defines one or more        }
  { containers that are required for each transaction type                     }
  { (ie: GET, GETDEFAULT...).  If the spec shows multiple options, check if a  }
  { ONEVALUE can be used else use one of the others.  If only one is noted, it }
  { must be used.  An example is ICAP_COMPRESSION which may only have one      }
  { attribute (TWCP_NONE) and must use enumeration as a container              }
  ConType := GetPrivateProfileInt('ConType', Container, 99, TwackerIni);
  if (StrPos(Container, ',') <> Nil)  { There is more than one option          }
  then begin
       if (StrPos(Capabilities, 'to') <> Nil)
       then ConType := TWON_RANGE
       else if (StrPos(Capabilities, '[') <> Nil)
	    then ConType := TWON_ARRAY
	    else if (StrPos(Capabilities, ',') <> Nil)
        	 then ConType := TWON_ENUMERATION
                 else ConType := TWON_ONEVALUE;
  end;

  { Pack up the containers and return the data                                 }
  case ConType of
  TWON_ONEVALUE    : begin
                       if IsCharAlpha(Capabilities[0])
            	       then begin
                            wValue := GetPrivateProfileInt(CapStr, Capabilities, 99, TwackerIni);
                            pValue := @wValue;
                       end
                       else begin
			    if (ValueType = TWTY_FIX32)
			    then begin
			         fix32  := FloatToFIX32(atof(Capabilities));
			         lValue := fix32.Long;
                                 pValue := @lValue;
			    end
			    else begin
                                 lValue := Atoi(Capabilities);
                                 pValue := @lValue;
                            end;
                       end;
                       BuildUpOneValue(pTW_CAPABILITY(pData), ValueType, pValue, Msg);
                     end;
  TWON_ENUMERATION : begin
                       StrTok(Capabilities, ',', Token);
                       twEnum.NumItems := 0;
                       while (StrLen(Token) > 0)
                       do begin
			  if IsCharAlpha(Token[0])
			  then Types[twEnum.NumItems] := GetPrivateProfileInt(CapStr, Token, 99, TwackerIni)
                          else Types[twEnum.NumItems] := Atoi(Token);
                          StrTok(Nil, ',', Token);
                          Inc(twEnum.NumItems);
                       end;
                       twEnum.ItemType     := ValueType;
                       twEnum.CurrentIndex := GetIndex(MSG_GETCURRENT, CapStr);
                       twEnum.DefaultIndex := GetIndex(MSG_GETDEFAULT, CapStr);
                       BuildUpEnumerationType(pTW_CAPABILITY(pData),
                                              @twEnum, @Types, Msg);
                     end;
  TWON_RANGE       : begin
		       twRange.ItemType := ValueType;
		       twRange.MinValue := RangeMinimum(Capabilities, twRange.ItemType);
		       twRange.MaxValue := RangeMaximum(Capabilities, twRange.ItemType);
		       twRange.StepSize := RangeStep(Capabilities, twRange.ItemType);
		       twRange.DefaultValue := 0;
		       twRange.CurrentValue := 0;
		       BuildUpRangeType(pTW_CAPABILITY(pData), @twRange, Msg);
                     end;
  TWON_ARRAY       : begin
		       lstrcpy(Capabilities, @Capabilities[1]);
		       Capabilities[lstrlen(Capabilities)-1] := #0;

                       StrTok(Capabilities, ',', Token);
		       twArray.NumItems := 0;
		       twArray.ItemType := ValueType;

		       while (StrLen(Token) <> 0)
		       do begin
			  if (IsCharAlpha(Token[0]))
			  then Types[twArray.NumItems] := GetPrivateProfileInt(CapStr, Token, 99, TwackerIni)
			  else Types[twArray.NumItems] := Atoi(Token);
			  StrTok(Nil, ',', Token);
			  inc(twArray.NumItems);
		       end;
                       BuildUpArrayType(pTW_CAPABILITY(pData), @twArray, @Types, Msg);
                     end;
  end;
end; { End SetValues.                                                          }


{------------------------------------------------------------------------------}
{ GetIndex -- Determines the position of the current or default attribute      }
{ within the scope of the ItemList for the Capabilities                        }
{------------------------------------------------------------------------------}

function GetIndex(Msg    : integer;
                  CapStr : PChar) : integer;
var TmpStr   : array[0..10] of char;
    index    : array[0..80] of char;
    Buffer   : array[0..256] of char;
    Position : integer;
    Token    : array[0..256] of char;
begin
  Position := 0;

  if (Msg = MSG_GETDEFAULT)
  then lstrcpyn(TmpStr, 'default', SizeOf(TmpStr))
  else lstrcpyn(TmpStr, 'current', SizeOf(TmpStr));

  GetPrivateProfileString(CapStr, 'Allowed', '',
                          Buffer, SizeOf(Buffer), SourceControl);
  GetPrivateProfileString(CapStr, TmpStr,    '',
                          index,  SizeOf(index),  SourceControl);

  StrTok(Buffer, ',', Token);
  while (StrLen(Token) <> 0)
  do begin
     if (StrLComp(index, Token, 8) = 0)
     then begin
          GetIndex := Position;
          Exit;
     end;
     StrTok(Nil, ',', Token);
     Inc(Position);
  end;

  GetIndex := 0;
end; { End GetIndex.                                                           }


{------------------------------------------------------------------------------}
{ SetNewCap -- Sets the current Capability to the user's selection.  Will      }
{ not return success if the item does not exist or can not be set.  If the     }
{ container is ONEVALUE or ARRAY, the current values are set, otherwise the    }
{ allowed values will be set.  Note that there is no validation when setting   }
{ the allowed values.                                                          }
{------------------------------------------------------------------------------}

function SetNewCap(pData  : pTW_CAPABILITY;
                   CapStr : PChar) : TW_INT16;
var Container    : array[0..40] of char;
    twacker      : array[0..MAX_TWPATH] of char;
    Attribute    : array[0..MAX_TWPATH] of char;
    TmpStr       : array[0..MAX_TWPATH] of char;
    Current      : array[0..20] of char;
    def          : array[0..20] of char;
    pOneValue    : pTW_ONEVALUE;
    pEnumValue   : pTW_ENUMERATION;
    pRangeValue  : pTW_RANGE;
    pArrayValue  : pTW_ARRAY;
    CapNum       : TW_INT16;
    index        : TW_UINT16;
    Buffer       : array[0..80] of char;
    found        : bool;
    Token        : array[0..256] of char;
    fix          : TW_FIX32;
    fix1         : TW_FIX32;
    fix2         : TW_FIX32;
    fix3         : TW_FIX32;
begin
  CapNum := 0;
  found  := False;
  GetWindowsDirectory(twacker, SizeOf(twacker));
  lstrcat(twacker, '\twacker.ini');

  { Check if this Cap can be set at all.  There may be multiple Caps, so check }
  { each.                                                                      }
  GetPrivateProfileString(CapStr, 'Set', '',
                          Container, SizeOf(Container), SourceControl);
  StrTok(Container, ',', Token);
  while (StrLen(Token) <> 0)
  do begin
     Str(pData^.ConType, Buffer);

     if (GetPrivateProfileInt('ConType', Token, 99, twacker) = TW_UINT16(pData^.ConType))
     then Found := True;
     StrTok(Nil, ',', Token);
  end;

  if Not(Found)
  then begin
       SetNewCap := TWCC_BADVALUE;
       Exit;
  end;

  case pData^.ConType of
  TWON_ONEVALUE    : begin
		       { check if the new Cap is in the allowed list.          }
		       pOneValue := pTW_ONEVALUE(GlobalLock(pData^.hContainer));
		       CapNum := TW_INT16(pOneValue^.Item);

		       { Doesn't check for validity of floats.                 }
		       if (pOneValue^.ItemType = TWTY_FIX32)
		       then begin
		            fix := pTW_FIX32(@pOneValue^.Item)^;
		            Str(FIX32ToFloat(fix):0:6, Attribute);
		       end
		       else begin
		            if Not(IsCapAllowed(CapNum, CapStr, Attribute))
		            then begin
			         MessageBox(0, 'IsCapAllowed', 'TWCC_BADVALUE', MB_OK);
				 SetNewCap := TWCC_BADVALUE;
                                 Exit;
			    end;
		       end;
		       GlobalUnlock(pData^.hContainer);

		       lstrcpyn(destination, 'Current', SizeOf(destination));
                     end;
  TWON_ENUMERATION : begin
		       FillChar(Attribute, SizeOf(Attribute), #0);
		       pEnumValue := pTW_ENUMERATION(GlobalLock(pData^.hContainer));

		       { put the items list together into a comma delimited    }
                       { string.                                               }
		       for index := 0 to (pEnumValue^.NumItems - 1)
		       do begin
			  FillChar(Buffer, SizeOf(Buffer), #0);
			  ConvertAttributeToText(pEnumValue^.ItemList[index], Buffer, CapStr);
			  lstrcat(Attribute, Buffer);
			  lstrcat(Attribute, ',');
		       end;

		       { Remove last comma.                                    }
		       Attribute[lstrlen(Attribute)-1] := #0;
		       GlobalUnlock(pData^.hContainer);
		       lstrcpyn(destination, 'Allowed', SizeOf(destination));
                     end;
  TWON_RANGE       : begin
		       pRangeValue := pTW_RANGE(GlobalLock(pData^.hContainer));

                       { build the string of the format min_to_max, step.      }
		       if (pRangeValue^.ItemType = TWTY_FIX32)
		       then begin
			    fix1   := pTW_FIX32(@pRangeValue^.MinValue)^;
			    fix2   := pTW_FIX32(@pRangeValue^.MaxValue)^;
			    fix3   := pTW_FIX32(@pRangeValue^.StepSize)^;

                            Str(FIX32ToFloat(fix1):0:6, Attribute);
                            StrCat(Attribute, '_to_');
                            Str(FIX32ToFloat(fix2):0:6, TmpStr);
                            StrCat(Attribute, TmpStr);
                            StrCat(Attribute, ', ');
                            Str(FIX32ToFloat(fix3):0:6, TmpStr);
                            StrCat(Attribute, TmpStr);

			    fix1 := pTW_FIX32(@pRangeValue^.CurrentValue)^;
			    Str(FIX32ToFloat(fix1):0:6, Current);

			    fix1 := pTW_FIX32(@pRangeValue^.DefaultValue)^;
			    Str(FIX32ToFloat(fix1):0:6, Def);
		       end
		       else begin
                            Str(pRangeValue^.MinValue, Attribute);
                            StrCat(Attribute, '_to_');
                            Str(pRangeValue^.MaxValue, TmpStr);
                            StrCat(Attribute, TmpStr);
                            StrCat(Attribute, ', ');
                            Str(pRangeValue^.StepSize, TmpStr);
                            StrCat(Attribute, TmpStr);

			    Str(pRangeValue^.CurrentValue, Current);
			    Str(pRangeValue^.DefaultValue, Def);
		       end;

		       { store the Current and default values                  }
		       WritePrivateProfileString(CapStr,
                                                 'Current',
                                                 Current,
                                                 SourceControl);
		       WritePrivateProfileString(CapStr,
                                                 'Default',
                                                 Def,
                                                 SourceControl);

		       GlobalUnlock(pData^.hContainer);
		       lstrcpyn(destination, 'Allowed', SizeOf(destination));

                     end;
  TWON_ARRAY       : begin
		       pArrayValue := pTW_ARRAY(GlobalLock(pData^.hContainer));

		       { build the string format [x, y, z...]                  }
		       FillChar(Attribute, SizeOf(Attribute), #0);
		       lstrcat(Attribute, '[');
		       for index := 0 to (pArrayValue^.NumItems - 1)
		       do begin
			  Str(pArrayValue^.ItemList[index], Buffer);
			  lstrcat(Attribute, Buffer);
			  lstrcat(Attribute, ',');
		       end;
		       { remove last comma.                                    }
		       Attribute[lstrlen(Attribute)-1] := #0;

                       lstrcat(Attribute, ']');

		       GlobalUnlock(pData^.hContainer);
		       lstrcpyn(destination, 'Allowed', SizeOf(destination));
		     end;
  end;

  { set Current to the new Cap.                                                }
  WritePrivateProfileString(CapStr, destination, Attribute, SourceControl);

  SetNewCap := TWRC_SUCCESS;
end; { End SetNewCap.                                                          }


{------------------------------------------------------------------------------}
{ IsCapAllowed -- Checks that the Attribute to set is in the Allowed list      }
{ either numerical or enumerated text                                          }
{------------------------------------------------------------------------------}

function IsCapAllowed(Cap       : TW_INT16;
                      CapStr    : PChar;
                      Attribute : PChar) : bool;
var Token   : array[0..256] of char;
    Buf     : array[0..30] of char;
    Buffer  : array[0..512] of char;
    TmpStr  : array[0..16] of char;
    IniFile : array[0..120] of char;
begin
  { make the inifile path.                                                     }
  GetWindowsDirectory(IniFile, SizeOf(IniFile));
  lstrcat(IniFile, '\twacker.ini');
  GetPrivateProfileString(CapStr,
                          'Allowed',
                          '',
                          Buffer,
                          SizeOf(Buffer),
                          SourceControl);

  { check if within range and on step boundry                                  }
  { if not on the step boundry, then CHECKSTATUS should be returned after the  }
  { value on the nearest step is set                                           }
  if (StrPos(Buffer, 'to') <> Nil)
  then begin
       { FIX32 change - the second parameter to the Range is hardcoded for now }
       if (Cap >= TW_INT16(RangeMinimum(Buffer, TWTY_UINT32))) and
          (Cap <= TW_INT16(RangeMaximum(Buffer, TWTY_UINT32)))
       then begin
            if (Cap mod RangeStep(Buffer, TWTY_UINT32) = 0)
	    then begin
                 Str(Cap, TmpStr);
                 StrCopy(Attribute, TmpStr);
            end
	    else begin
	         Cap := Cap div TW_INT16(RangeStep(Buffer, TWTY_UINT32));
		 Cap := Cap * TW_INT16(RangeStep(Buffer, TWTY_UINT32));
                 Str(Cap, TmpStr);
                 StrCopy(Attribute, TmpStr);
	    end;
	    IsCapAllowed := True;
            Exit;
       end;
       IsCapAllowed := False;
       Exit;
  end
  { make array look like enumerated list (remove brackets)                     }
  else begin
       if (StrPos(Buffer, '[') <> Nil)
       then FillChar(Buffer, lstrlen(Buffer) - 1, Buffer[1]);
  end;

  { check if in comma delimited list or a onevalue.                            }
  StrTok(Buffer, ',', Token);
  while (StrLen(Token) <> 0)
  do begin
     lstrcpyn(Buf, Token, SizeOf(Buf));
     if IsCharAlpha(Buf[0])
     then begin
          if (Cap = TW_INT16(GetPrivateProfileInt(CapStr, Buf, 99, IniFile)))
	  then begin
	       lstrcpyn(Attribute, Buf, SizeOf(Buf));
               IsCapAllowed := True;
               Exit;
	  end;
     end
     else begin
          if (Cap = TW_INT16(atoi(Buf)))
	  then begin
	       lstrcpyn(Attribute, Buf, SizeOf(Buf));
               IsCapAllowed := True;
               Exit;
	  end;
     end;
     StrTok(Nil, ',', Token);
  end;
  IsCapAllowed := False;
end; { End IsCapAllowed.                                                       }


{------------------------------------------------------------------------------}
{ WithinRange --                                                               }
{------------------------------------------------------------------------------}

function WithinRange(Cap    : TW_UINT16;
                     CapStr : PChar) : bool;
begin
  WithinRange := True;
end; { WithinRange.                                                            }


{------------------------------------------------------------------------------}
{ ConvertAttributeToText --                                                    }
{------------------------------------------------------------------------------}

procedure ConvertAttributeToText(Attrib  : TW_UINT32;
                                 StrText : PChar;
                                 CapStr  : PChar);
var Buffer  : array[0..2048] of char;
    Token   : array[0..30] of char;
    x, y    : integer;
    IniFile : array[0..120] of char;
    TmpStr  : array[0..16] of char;
begin
  { make the inifile path.                                                     }
  GetWindowsDirectory(IniFile, SizeOf(IniFile));
  lstrcat(IniFile, '\twacker.ini');

  { the 16 sdk does not support the section command.                           }
  {$IFDEF WIN32}
    GetPrivateProfileSection(CapStr, Buffer, SizeOf(Buffer), IniFile);
  {$ELSE}
    GetPrivateProfileSection16(CapStr, Buffer, SizeOf(Buffer));
  {$ENDIF}

  { if the Cap is numerical,  there will be no section entry for it in         }
  { the .ini file.  Return the number.                                         }
  if (Buffer[0] = #0)
  then begin
       Str(Attrib, TmpStr);
       StrCopy(StrText, TmpStr);
  end;

  x := 0;
  FillChar(Token, SizeOf(Token), #0);
  while(Buffer[x] <> #0)
  do begin
     y := 0;
     FillChar(Token, 30, #0);
     while(Buffer[x] <> '=')
     do begin
        Token[y] := Buffer[x];
	inc(x);
        inc(y);
     end;
     inc(x);

     if (attrib = TW_UINT32(atoi(@Buffer[x])))
     then begin
          lstrcpyn(StrText, Token, SizeOf(Token));
	  FillChar(Token, SizeOf(Token), #0);
	  Exit;
     end;
     while(Buffer[x] <> #0)
     do inc(x);
     inc(x);
  end;
end; { End ConvertAttributeToText.                                             }


{------------------------------------------------------------------------------}
{ RangeMinimum -                                                               }
{------------------------------------------------------------------------------}

function RangeMinimum(Attr     : PChar;
                      ItemType : TW_UINT16) : TW_UINT32;
var Token   : array[0..256] of char;
    TmpCopy : array[0..MAX_TWPATH] of char;
    value   : TW_UINT32;
    fix32   : TW_FIX32;
begin
  lstrcpy(TmpCopy, attr);
  StrTok(TmpCopy, '_', Token);

  if (ItemType = TWTY_FIX32)
  then begin
       fix32 := FloatToFIX32(atof(Token));

       Value := TW_UINT32(fix32.Long);
       RangeMinimum := Value;
  end
  else RangeMinimum := TW_UINT32(atoi(Token));
end; { End RangeMinimum.                                                       }


{------------------------------------------------------------------------------}
{ RangeMaximum                                                                 }
{------------------------------------------------------------------------------}

function RangeMaximum(Attr     : PChar;
                      ItemType : TW_UINT16) : TW_UINT32;
var Temp      : PChar;
    TmpCopy   : array[0..MAX_TWPATH] of char;
    TmpResult : array[0..MAX_TWPATH] of char;
    Value     : TW_UINT32;
    fix32     : TW_FIX32;
begin
  lstrcpy(TmpCopy, Attr);
  Temp := StrPos(TmpCopy, 'to_');
  Temp := StrTok(Temp, ',', TmpResult) + 3;

  if (ItemType = TWTY_FIX32)
  then begin
       fix32 := FloatToFIX32(atof(Temp));
       Value := TW_UINT32(fix32.Long);
       RangeMaximum := Value;
  end
  else RangeMaximum := TW_UINT32(atoi(Temp));
end; { End RangeMaximum.                                                       }


{------------------------------------------------------------------------------}
{ RangeStep                                                                    }
{------------------------------------------------------------------------------}

function RangeStep(Attr     : PChar;
                   ItemType : TW_UINT16) : TW_UINT32;
var Temp    : PChar;
    TmpCopy : array[0..MAX_TWPATH] of char;
    Value   : TW_UINT32;
    fix32   : TW_FIX32;
begin
  lstrcpy(TmpCopy, Attr);
  Temp := StrPos(TmpCopy, ',') + 1;

  if (ItemType = TWTY_FIX32)
  then begin
       fix32 := FloatToFIX32(atof(temp));
       Value := TW_UINT32(fix32.Long);
       RangeStep := Value;
  end
  else RangeStep := TW_UINT32(atoi(temp));
end; { End RangeStep.                                                          }


{$IFNDEF WIN32}

{------------------------------------------------------------------------------}
{ GetPrivateProfileSection16 -- Performs GetPrivateProfileSection              }
{ for a 16 bit application.  The function is not defined in the                }
{ 16 bit Win API.                                                              }
{------------------------------------------------------------------------------}

procedure GetPrivateProfileSection16(Cap     : PChar;
                                     LineOut : PChar;
                                     Length  : integer);
var Buffer  : array[0..512] of char;
    Entry   : array[0..256] of char;
    Token   : array[0..40] of char;
    x       : integer;
    LinePos : integer;
    Line    : array[0..40] of char;
    IniFile : array[0..MAX_TWPATH] of char;
begin
  LinePos := 0;
  GetWindowsDirectory(IniFile,  SizeOf(IniFile));
  lstrcat(IniFile, '\twacker.ini');

  GetPrivateProfileString(Cap, Nil, '', Buffer, SizeOf(Buffer), IniFile);
  for x := 0 to (SizeOf(Buffer) - 1)
  do begin
     if (Buffer[x] = #0)
     then begin
          Buffer[x] := ',';
          if (Buffer[x+1] = #0)
          then begin
               Buffer[x] := #0;
               Exit;
          end;
      end;
  end;

  FillChar(LineOut, Length, #0);

  StrTok(Buffer, ',', Entry);
  while (StrLen(Entry) <> 0)
  do begin
     FillChar(Token, SizeOf(Token), #0);
     GetPrivateProfileString(Cap, Entry, '', Token, SizeOf(Token), IniFile);
     lstrcpyn(Line, Entry, SizeOf(line));
     lstrcat(Line, '=');
     lstrcat(Line, Token);
     StrTok(Nil, ',', Entry);
     StrLCat(@LineOut[LinePos], Line, lstrlen(line));
     { memcpy(dummy, LineOut, len);                                            }
     LinePos := LinePos + lstrlen(Line) + 1;
  end;
end; { End GetPrivateProfileSection16.                                         }

{$ENDIF}


{------------------------------------------------------------------------------}
{ FloatToFIX32 -- Convert a floating point value into a FIX32.                 }
{------------------------------------------------------------------------------}


function FloatToFIX32(floater : real) : TW_FIX32;
var Fix32_value : TW_FIX32;
    Value       : TW_INT32;
begin
  Value := TW_INT32(Round(floater * 65536.0 + 0.50));
  Fix32_value.Whole := LOWORD(value shr 16);
  Fix32_value.Frac  := LOWORD(value and $0000FFFF);
  {$IFDEF WIN32}
    FloatToFIX32.Long := Fix32_value.Long;
  {$ELSE}
    FloatToFIX32^   := Fix32_value;
  {$ENDIF}
end; { End FloatToFIX32.                                                       }


{------------------------------------------------------------------------------}
{ FIX32ToFloat -- Convert a FIX32 value into a floating point value            }
{------------------------------------------------------------------------------}

function FIX32ToFloat(fix32 : TW_FIX32) : real;
var floater : real;
begin
  floater      := 1.0 * fix32.Whole + (1.0 * fix32.Frac / 65536.0);
  FIX32ToFloat := floater;
end; { End FIX32ToFloat.                                                       }


end. { End Unit twd_Caps.                                                      }
